
Object subclass: #ActivateProgramGenerator
  instanceVariableNames: 
    ' programName switches states deltaStates workingSet caseMap unbreakableCasesList '
  classVariableNames: ''
  poolDictionaries: ''
  category: 'JOOP'
!

Object subclass: #State
  instanceVariableNames: 
    ' name onSwitchNames offSwitchNames child '
  classVariableNames: ''
  poolDictionaries: ''
  category: 'JOOP'
!

Object subclass: #Switch
  instanceVariableNames: 
    ' name onCode offCode '
  classVariableNames: ''
  poolDictionaries: ''
  category: 'JOOP'
!


!ActivateProgramGenerator class methodsFor: 'instance creation'!
 
new
	^super new initialize
! !

!ActivateProgramGenerator class methodsFor: 'program generation'!
   
generateProgram: name stateToSwitchMap: triples1 switchToCodeMap: triples2
	| generator |
	generator := self new programName: name.
	triples1 groupsDo: [:stateName :onSwitchNames :offSwitchNames |
		generator addState: (State new
			name: stateName;
			onSwitchNames: onSwitchNames;
			offSwitchNames: offSwitchNames)].
	triples2 groupsDo: [:switchName :onCode :offCode |
		generator addSwitch: (Switch new
			name: switchName;
			onCode: onCode;
			offCode: offCode)].
	generator prepareForCodeGeneration.
	^generator generateCode
! !

!ActivateProgramGenerator class methodsFor: 'examples'!
   
example1
	"ActivateProgramGenerator example1"
	^self generateProgram: 'activate'
		stateToSwitchMap: #(
			"(stateName (... on switches ...) (... off switches ...))"
			(flatColor () (texturing lighting depthTest stippling blending tracking))
			(flatDashes (stippling) (texturing lighting depthTest blending tracking))
			(flatBlend (blending) (texturing lighting depthTest stippling tracking))
			(litColor (lighting tracking) (texturing depthTest stippling blending)))
		switchToCodeMap: #(
			"(switchName 'code to turn on' 'code to turn off')"
			(texturing 'glEnable (GL_TEXTURE_2D);' 'glDisable (GL_TEXTURE_2D);')
			(lighting 'glEnable (GL_LIGHTING);' 'glDisable (GL_LIGHTING);')
			(depthTest 'glEnable (GL_DEPTH_TEST);' 'glDisable (GL_DEPTH_TEST);')
			(stippling 'glEnable (GL_LINE_STIPPLE);' 'glDisable (GL_LINE_STIPPLE);')
			(blending 'glEnable (GL_BLEND); glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);' 'glDisable (GL_BLEND);')
			(tracking 'glEnable (GL_COLOR_MATERIAL);' 'glDisable (GL_COLOR_MATERIAL);'))
! !


!ActivateProgramGenerator methodsFor: 'initializing/finalizing'!
 
initialize
	switches := Dictionary new.
	states := Dictionary new.
! !

!ActivateProgramGenerator methodsFor: 'get/set'!
   
programName: aString
	programName := aString
!

programName
	^programName
! !

!ActivateProgramGenerator methodsFor: 'adding'!
  
addSwitch: aSwitch
	switches at: aSwitch name put: aSwitch
!
  
addState: aState
	states at: aState name put: aState
! !

!ActivateProgramGenerator methodsFor: 'querying'!
 
stateNames
	^states keys asSortedCollection asArray
! !

!ActivateProgramGenerator methodsFor: 'printing'!
  
printOn: aStream
	aStream cr; << 'Generator switches:'.
	switches do: [:switch | aStream cr; << switch].
	aStream cr; << 'Generator states:'.
	states do: [:state | aStream cr; << state].
! !

!ActivateProgramGenerator methodsFor: 'state deltas'!

computeDeltaStates
	| state |
	deltaStates := Dictionary new.
	states do: [:state1 |
		states do: [:state2 |
			state := state1 deltaToGet: state2.
			deltaStates at: state name put: state]].
! !

!ActivateProgramGenerator methodsFor: 'case lists'!
   
casesDo: aBlock
	"self casesDo: [:state | ...]"
	unbreakableCasesList do: [:caseList |
		caseList do: aBlock]
!
 
minimalState
	| state |
	state := workingSet any.
	workingSet do: [:anotherState |
		anotherState size < state size ifTrue: [
			 state := anotherState]].
	^state
!
 
differenceCasesDo: aBlock
	"self differenceCasesDo: [:state :differenceState :needBreak | ...]"
	| state1 state2 |
	unbreakableCasesList do: [:caseList |
		"CaseList cannot be empty."
		1 to: caseList size - 1 do: [:index |
			state1 := caseList at: index.
			state2 := caseList at: index+1.
			aBlock value: state1 value: (state1 subtract: state2) value: false].
		"The last superstate has no difference."
		aBlock value: (caseList last) value: (caseList last) value: true]
!
 
minimalChildlessSuperstateOf: state1
	"Returns a superset of state1 that as yet has no child."
	| candidate |
	candidate := nil.
	workingSet do: [:state2 |
		(state2 child isNil and: [state1 isStrictSubsetOf: state2]) ifTrue: [
			(candidate isNil or: [candidate size > state2 size]) ifTrue: [
				candidate := state2]]].
	^candidate
!

computeUnbreakableCasesList
	"Compute an alternative to deltaStates which
	has been partitioned into lists representing successive
	cases without an intervening break. Each state in
	a list represents successive proper subsets."
	| state superstate next cases |
	"Create a parent/child ordering from superset to subset states."
	unbreakableCasesList := OrderedCollection new.
	workingSet := deltaStates asSet asOrderedCollection.
	[workingSet isEmpty] whileFalse: [
		"Choose any minimal sized state."
		state := self minimalState.
		workingSet remove: state.
		"Choose a state that can be a superstate."
		superstate := self minimalChildlessSuperstateOf: state.
		superstate isNil
			ifTrue: [
				cases := OrderedCollection new.
				next := state.
				[next notNil] whileTrue: [
					cases add: next.
					next := next child].
				unbreakableCasesList add: cases]
			ifFalse: [superstate child: state]].
! !

!ActivateProgramGenerator methodsFor: 'case map'!
  
namesMappingTo: aNumber
	"The (fromState, toState) pairs as a string."
	| stream |
	stream := WriteStream on: (String new: 100).
	caseMap keysAndValuesDo: [:key :number |
		number = aNumber ifTrue: [
			stream position = 0 ifFalse: [stream << ', '].
			stream << '['; << key first; << '=>'; << key last; << ']']].
	^stream contents
!
  
computeCaseMap
	"Maps each delta state to a number between 0 and the
	number of unique delta states."
	| used count index |
	caseMap := Dictionary new.
	used := OrderedCollection new.

	"First, handle all unique states."
	count := -1.
	self casesDo: [:state |
		caseMap at: state name put: (count := count + 1).
		used add: state].

	"Next, those equal to one of above."
	deltaStates keysAndValuesDo: [:key :state |
		(caseMap includesKey: key) ifFalse: [
			index := (used indexOf: state) - 1.
			caseMap at: key put: index]]
! !

!ActivateProgramGenerator methodsFor: 'code generation'!

generateActivateFunctionInto: stream
	| caseNumber break last |
	stream cr; << 'void '; << self programName; << ' (State newState) {'.
	stream cr; tab; << 'switch (map [oldState] [newState]) {'.
	break := (ReadWriteStream on: '') cr; tab; tab; tab; << 'break;'; contents.
	last := ''.
	self differenceCasesDo: [:state :differenceState :needBreak |
		caseNumber := caseMap at: state name.
		stream << last.
		stream cr; tab; tab; << 'case '; << caseNumber; << ': //'.
		stream << (self namesMappingTo: caseNumber).
		stream cr; tab; tab; tab;
			<< '//on '; << (state onSwitchNames);
			<< ' off '; << (state offSwitchNames).
		self generateCodeFor: differenceState into: stream lineTabs: 3.
		last := needBreak ifTrue: [break] ifFalse: ['']].
	stream cr; tab; << '}'.
	stream cr; tab; << 'oldState = newState;'.
	stream cr; << '}'.
!

generateStateEnumerationInto: stream
	| last |
	stream cr; << 'enum State '.
	last := '{'.
	self stateNames do: [:name |
		stream << last << name.
		last := ', '].
	stream << '};'.
!
  
generateInitializeFunctionInto: stream
	| state |
	stream cr; << 'static State oldState;'.
	stream cr; << 'void initialize';
		<< self programName asCapitalizedString; << ' () {'.
	state := states any.
	self generateCodeFor: state into: stream lineTabs: 1.
	stream cr; tab; << 'oldState = '; << state name; << ';'.
	stream cr; << '}'.
!
   
generateCodeFor: state into: stream lineTabs: count
	state onSwitchNames do: [:name |
		stream cr; tab: count; << (switches at: name) onCode].
	state offSwitchNames do: [:name |
		stream cr; tab: count; << (switches at: name) offCode].
!
  
prepareForCodeGeneration
	self computeDeltaStates.
	self computeUnbreakableCasesList.
	self computeCaseMap
!

generateStateToCaseMappingInto: stream
	| names lastName lastRow lastElement |
	names := self stateNames.
	lastName := names last.
	stream cr; << 'long map ['; << lastName; << ']['; << lastName; << '] = {'.
	lastRow := ''.
	names do: [:oldName |
		stream << lastRow; cr; tab.
		lastElement := '{'.
		names do: [:newName |
			stream << lastElement << (caseMap at: (Array with: oldName with: newName)).
			lastElement := ', '].
		stream << '}'.
		lastRow := ','].
	stream cr; << '};'
!
  
generateCode
	| stream |
	stream := WriteStream on: (String new: 1000).
	self generateStateEnumerationInto: stream.
	self generateStateToCaseMappingInto: stream.
	self generateInitializeFunctionInto: stream.
	self generateActivateFunctionInto: stream.
	^stream contents
! !
ActivateProgramGenerator comment: ''!


!State class methodsFor: 'comment'!
   
comment
	"
	The fundamental contributions provided by this class are methods that perform
	boolean-like operations with another state; i.e., 'isStrictSubsetOf: aState',
	'subtract: aState', and 'deltaToGet: state.'

	More specifically, state1 is a strict subset of state2 if the two states are not
	equal and every on switch name in state1 is also in state2 (the same requirement
	also applies to the off switch names). In that situation, 'state1 subtract: state2'
	gives us the difference between the two; i.e., the state difference.

	Even if there is no relationship between state1 and state2, we can still compute
	a delta state ('state1 deltaToGet: state2'), i.e., a new state that describes
	what has to be changed to go from state1 to state2; e.g., which switches have
	to be turned on and which have to be turned off (anything that does not change
	is not mentioned). This is subtly different from subtracting. Although we can
	subtract the 'lighting' switch from two states if 'lighting' is on in both states or
	off in both states, we can't subtract it if it is on in one state and off in the other.
	In other words, subtracting makes sense for strict subsets but not for arbitrary
	states."
! !


!State methodsFor: 'get/set'!

child
	^child
!
   
name: aSymbol
	name := aSymbol
!
  
name
	^name
!
 
offSwitchNames: aCollection
	offSwitchNames := aCollection
!
  
child: aState
	child := aState
!
  
onSwitchNames
	^onSwitchNames
!
   
offSwitchNames
	^offSwitchNames
!
 
onSwitchNames: aCollection
	onSwitchNames := aCollection
! !

!State methodsFor: 'printing'!

printOn: aStream
	aStream
		cr;
		<< name; cr; tab;
		<< 'on: '; << (self onSwitchNames); cr; tab;
		<< 'off: '; << (self offSwitchNames)
! !

!State methodsFor: 'comparing'!
  
= anObject
	(anObject isKindOf: State)
		ifFalse: [^false].
	self onSwitchNames asSortedCollection
		= anObject onSwitchNames asSortedCollection
		ifFalse: [^false].
	^self offSwitchNames asSortedCollection
		= anObject offSwitchNames asSortedCollection
!
 
hash
	^self onSwitchNames size * 100 + self offSwitchNames size
!
 
isStrictSubsetOf: aState
	(aState onSwitchNames
		includesAll: self onSwitchNames)
			ifFalse: [^false].
	(aState offSwitchNames
		includesAll: self offSwitchNames)
			ifFalse: [^false].
	^(self onSwitchNames size
		< aState onSwitchNames size) or: [
	(self offSwitchNames size
		< aState offSwitchNames size)]
!
 
size
	^self onSwitchNames size + self offSwitchNames size
! !

!State methodsFor: 'state deltas'!
   
deltaToGet: state
	"A delta state (i.e., just the changes) needed to convert 'self' to 'state'.
	The new state's name is the ['fromState', 'toState'] pair."
	| state1 state2 turnOns turnOffs |
	state1 := self. state2 := state. "easier to understand"
	turnOns := OrderedCollection new.
	turnOffs := OrderedCollection new.
	state2 onSwitchNames do: [:name |
		(state1 onSwitchNames includes: name) ifTrue: [
			"no change needed"] ifFalse: [
		(state1 offSwitchNames includes: name) ifTrue: [
			turnOns add: name] ifFalse: [
		"Must have been don't care for state2. But state1 cares."
			turnOns add: name]]].
	state2 offSwitchNames do: [:name |
		(state1 offSwitchNames includes: name) ifTrue: [
			"no change needed"] ifFalse: [
		(state1 onSwitchNames includes: name) ifTrue: [
			turnOffs add: name] ifFalse: [
		"Must have been don't care for state2. But state1 cares."
			turnOffs add: name]]].
	^State new
		name: (Array with: state1 name with: state2 name);
		onSwitchNames: turnOns asArray;
		offSwitchNames: turnOffs asArray
! !

!State methodsFor: 'state differences'!
   
subtract: aState
	"Assumes self is a superset of aState."
	aState isNil ifTrue: [^self].
	^self class new
		onSwitchNames: (self onSwitchNames shallowCopy asOrderedCollection
			removeAll: aState onSwitchNames; yourself);
		offSwitchNames: (self offSwitchNames shallowCopy asOrderedCollection
			removeAll: aState offSwitchNames; yourself)
! !
State comment: ''!



!Switch methodsFor: 'get/set'!
   
name
	^name
!
 
onCode
	^onCode
!
 
onCode: aString
	onCode := aString
!
  
offCode: aString
	offCode := aString
!

offCode
	^offCode
!
   
name: aSymbol
	name := aSymbol
! !

!Switch methodsFor: 'printing'!
 
printOn: aStream
	aStream
		<< name; cr; tab;
		<< 'on: '; << (self onCode); cr; tab;
		<< 'off: '; << (self offCode)
! !
Switch comment: ''!
